import shlex
import urllib.parse
import uuid
from os import path

import yaml
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.core.validators import MinLengthValidator
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils.functional import cached_property
from enumfields import Enum, EnumField
from rest_framework.exceptions import ValidationError

import pgtrigger
from arkindex.documents.models import Element
from arkindex.images.models import ImageServer
from arkindex.process.managers import ActivityManager, CorpusWorkerVersionManager
from arkindex.process.providers import get_provider
from arkindex.process.utils import get_default_farm_id
from arkindex.project.aws import S3FileMixin, S3FileStatus
from arkindex.project.fields import ArrayField, MD5HashField
from arkindex.project.models import IndexableModel
from ponos.models import Artifact, State, Workflow


class ActivityState(Enum):
    """
    Store the state of the workers activity tracking for a process.
    To support large elements set, the state is asynchronously set to `ready` after a process
    has been started and worker activities have been initialized on its elements.
    """
    Disabled = 'disabled'
    Pending = 'pending'
    Ready = 'ready'
    Error = 'error'


class ProcessMode(Enum):
    Files = 'files'
    Repository = 'repository'
    IIIF = 'iiif'
    Workers = 'workers'
    Transkribus = 'transkribus'
    Template = 'template'
    S3 = 's3'
    Training = 'training'


class Process(IndexableModel):
    """
    A single import workflow
    """

    name = models.CharField(null=True, blank=True, max_length=100)
    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='processes')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='processes', null=True)
    mode = EnumField(ProcessMode, max_length=30)
    files = models.ManyToManyField('process.DataFile', related_name='processes')
    revision = models.ForeignKey(
        'process.Revision', related_name='processes', on_delete=models.CASCADE, blank=True, null=True)
    workflow = models.OneToOneField('ponos.Workflow', on_delete=models.SET_NULL, null=True, blank=True)
    versions = models.ManyToManyField('process.WorkerVersion', through='process.WorkerRun', related_name='processes')
    activity_state = EnumField(ActivityState, max_length=32, default=ActivityState.Disabled)

    # Used to define the root element on which the workflow will run
    element = models.ForeignKey('documents.Element', on_delete=models.SET_NULL, null=True, blank=True)

    # Used to define the element type during folder elements creation for Images, PDF, IIIF and S3 imports
    folder_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='folder_imports',
    )

    # Used to define the element type during non-folder elements creation for Images, PDF, IIIF and S3 imports,
    # or to filter elements by type with Workers processes
    element_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='element_imports',
    )

    # Used to filter out elements with a name that doesn't contain the substring, only on Workers processes
    name_contains = models.CharField(null=True, blank=True, max_length=250)

    # Used to save a user's selection for Workers processes
    elements = models.ManyToManyField(
        'documents.Element',
        through='process.ProcessElement',
        related_name='processes',
    )

    # Used to include sub-elements of the actual query in Workers processes
    load_children = models.BooleanField(default=False)

    # Used to define the collection ID and entities import for Transkribus import
    collection_id = models.PositiveIntegerField(null=True, blank=True)

    # Use elements cache in Workers processes
    use_cache = models.BooleanField(default=False)

    # Allows to use a GPU in Workers processes
    use_gpu = models.BooleanField(default=False, blank=True)

    # Refers a workflow template
    template = models.ForeignKey(
        'process.Process',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='applications',
    )

    # S3 bucket name, and prefix within the bucket, for ProcessMode.S3
    # Bucket names are between 3 and 63 characters long
    bucket_name = models.CharField(
        max_length=63,
        blank=True,
        null=True,
        validators=[MinLengthValidator(3)],
    )
    # Object keys are up to 1024 characters long
    prefix = models.CharField(max_length=1024, blank=True, null=True)

    # Training specific fields
    model = models.ForeignKey(
        'training.Model',
        related_name='training_processes',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    train_folder = models.ForeignKey(
        'documents.Element',
        related_name='train_folder_processes',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    validation_folder = models.ForeignKey(
        'documents.Element',
        related_name='validation_folder_processes',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    test_folder = models.ForeignKey(
        'documents.Element',
        related_name='test_folder_processes',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    class Meta:
        ordering = ['corpus', '-created']
        verbose_name_plural = 'processes'

    @property
    def state(self):
        if not self.workflow:
            return State.Unscheduled
        # This allows annotating a Process queryset with "last_run" and preventing duplicate SQL queries
        if hasattr(self, 'last_run'):
            # last_run may be None when there is a workflow without any tasks
            if self.last_run is None:
                return State.Unscheduled
            return self.workflow.get_state(self.last_run)
        else:
            return self.workflow.state

    def _get_filters(self):
        filters = {
            "corpus_id": self.corpus_id,
        }
        if self.name_contains:
            filters['name__icontains'] = self.name_contains
        if self.element_type:
            filters['type_id'] = self.element_type_id

        return filters

    def copy_runs(self, new_process):
        """
        Copies this process' WorkerRuns to another process.

        This uses a dict to remap the existing parent UUIDs to the newly created UUIDs:
        otherwise, the runs on the other process would still have as their parents
        the runs of the current process, breaking their dependencies.
        """
        # Keep the relation between template_worker_runs and new_process_worker_runs ids
        new_runs = {}

        # First we copy each worker runs
        for run in self.worker_runs.all():
            # Create a new WorkerRun with same version, configuration and parents.
            new_run = WorkerRun(
                process=new_process,
                version_id=run.version_id,
                parents=run.parents,
                configuration_id=run.configuration_id,
                summary=run.summary
            )
            # Save the correspondence between this process' worker_run and the new one
            new_runs[run.id] = new_run

        # Remap parent ids correctly
        for run in new_runs.values():
            # If there are parents, we need to use the correct worker_run id, the newly created one
            run.parents = [new_runs[parent_id].id for parent_id in run.parents]
        WorkerRun.objects.bulk_create(new_runs.values())

    def list_elements(self):
        """
        Return a queryset of elements involved in this process
        """
        if self.mode != ProcessMode.Workers:
            return Element.objects.none()

        elements = None

        if self.element:
            # Assert the element has the same corpus as the process
            assert self.element.corpus_id == self.corpus_id, \
                f'Element {self.element.id} is not part of corpus {self.corpus_id}'
            # Handle a single element
            elements = Element.objects.filter(id=self.element.id)

        elif self.elements.exists():
            # Handle a selection of elements
            elements = self.elements.all()
            # Check all elements are in the same corpus as the process
            corpus = list(elements.values_list('corpus_id', flat=True).distinct())
            assert len(corpus) == 1 and corpus[0] == self.corpus_id, \
                f'Some elements on this process are not part of corpus {self.corpus_id}'

        if elements is not None and self.load_children:
            # Load all the children elements whose path contains the pre-selected elements
            # Those children are appended to the pre-selection
            elements |= Element.objects.filter(
                paths__path__overlap=map(str, elements.values_list('id', flat=True)),
            )

        # Load the full corpus, only when elements has not been populated before
        if elements is None:
            # Handle all elements of the process' corpus
            elements = Element.objects.filter(corpus=self.corpus_id)
            if not self.load_children:
                # Use top-level elements only
                elements = elements.filter(paths__path=[])

        # Filter elements depending on process properties
        return elements.filter(**self._get_filters())

    def build_workflow(self, farm=None, chunks=None, thumbnails=False, corpus_id=None, **kwargs):
        '''
        Create a ponos workflow with a recipe according to configuration
        '''
        ml_workflow_chunks = 1
        import_task_name = 'import'

        assert self.workflow is None, 'A workflow is already setup'

        if self.mode == ProcessMode.Repository:
            assert self.revision is not None, \
                'A revision is required to create an import workflow from GitLab repository'
            if not self.revision.repo.enabled:
                raise ValidationError('Git repository does not have any valid credentials')
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': (
                        f"python -m arkindex_tasks.import_git {self.revision.id}"
                        + (f' --corpus-id {corpus_id}' if corpus_id else '')
                    )
                },
            }

        elif self.mode == ProcessMode.IIIF:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': f'python -m arkindex_tasks.import_iiif.process {self.id}',
                },
            }

        elif self.mode == ProcessMode.Transkribus:
            import_task_name = 'import_arkindex'
            tasks = {
                'export_transkribus': {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': f'python -m arkindex_tasks.export_transkribus {self.collection_id}'
                },
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_transkribus '
                               f'--job-path /data/export_transkribus/transkribus_export_job.json '
                               f'--corpus {self.corpus.id}',
                    'parents': ['export_transkribus'],
                }
            }

        elif self.mode == ProcessMode.Files:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': f'python -m arkindex_tasks.import_files {self.id}',
                },
            }

        elif self.mode == ProcessMode.S3:
            assert self.bucket_name, 'Missing S3 bucket name'
            assert self.folder_type is not None, 'Missing folder type'
            assert self.element_type is not None, 'Missing element type'
            command = (
                'python -m arkindex_tasks.import_s3'
                f' --corpus={self.corpus_id}'
                f' --bucket={shlex.quote(self.bucket_name)}'
                f' --folder-type={shlex.quote(self.folder_type.slug)}'
                f' --page-type={shlex.quote(self.element_type.slug)}'
                f' --iiif-base-url={shlex.quote(ImageServer.objects.ingest.url)}'
            )
            if self.prefix:
                command += f' --prefix={shlex.quote(self.prefix)}'
            if self.element:
                command += f' --element={self.element_id}'
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': command,
                },
            }

        elif self.mode == ProcessMode.Training:
            worker_version = kwargs.get('worker_version')
            model_version = kwargs.get('model_version')
            worker_configuration = kwargs.get('worker_configuration')
            assert worker_version, 'A worker version is required to start the training'

            worker_run = WorkerRun.objects.create(
                process_id=self.id,
                version_id=worker_version.id,
                configuration_id=worker_configuration and worker_configuration.id,
                parents=[],
            )

            env = {
                'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
                'WORKER_VERSION_ID': str(worker_version.id),
            }
            if model_version:
                env['MODEL_VERSION_ID'] = str(model_version.id)

            # Build the training task, as there is no initial task
            tasks = {
                'training': {
                    'image': worker_version.docker_image_iid,
                    'command': worker_version.docker_command,
                    'artifact': str(worker_version.docker_image_id),
                    'env': env,
                    'requires_gpu': self.use_gpu,
                    'shm_size': worker_version.docker_shm_size,
                }
            }

        elif self.mode == ProcessMode.Workers:
            import_task_name = 'initialisation'
            if chunks is not None:
                assert chunks <= settings.MAX_CHUNKS, f'Import distribution is limited to {settings.MAX_CHUNKS} chunks'
                ml_workflow_chunks = chunks

            args = [
                'python', '-m', 'arkindex_tasks.init_elements', str(self.id),
                '--chunks-number', str(ml_workflow_chunks),
            ]
            if self.use_cache:
                args.append('--use-cache')

            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': ' '.join(args),
                }
            }

        else:
            raise NotImplementedError

        # Handle chunks and thumbnails generation on processes that list elements during the initial task
        # The S3 import does not generate a list of elements yet
        if self.mode in (
            ProcessMode.Workers,
            ProcessMode.Transkribus,
            ProcessMode.IIIF,
            ProcessMode.Files
        ):
            elts_chunk_files = ['elements.json']
            if ml_workflow_chunks > 1:
                elts_chunk_files = [f'elements_chunk_{n}.json' for n in range(1, ml_workflow_chunks + 1)]

            # Build up tasks for each worker run on each chunk
            worker_runs = list(self.worker_runs.select_related('version__worker__repository').using('default'))
            for index, elts_chunk in enumerate(elts_chunk_files, start=1):
                # Add a name suffix if task is distributed
                task_suffix = ''
                chunk = None
                if len(elts_chunk_files) > 1:
                    task_suffix = f'_{index}'
                    chunk = index

                elements_path = shlex.quote(path.join('/data', import_task_name, elts_chunk))

                # Generate thumbnails directly after import step
                if thumbnails:
                    tasks[f'thumbnails{task_suffix}'] = {
                        'image': settings.ARKINDEX_TASKS_IMAGE,
                        'command': f'python3 -m arkindex_tasks.generate_thumbnails {elements_path}',
                        'parents': [import_task_name],
                    }

                # Generate a task for each WorkerRun on the Process
                for worker_run in worker_runs:
                    task_name = f'{worker_run.version.slug}{task_suffix}'
                    task_recipe = worker_run.build_task_recipe(
                        import_task_name,
                        elements_path,
                        chunk=chunk,
                        workflow_runs=worker_runs,
                    )
                    if self.use_gpu:
                        if worker_run.version.gpu_usage in (WorkerVersionGPUUsage.Required, WorkerVersionGPUUsage.Supported):
                            task_recipe["requires_gpu"] = True
                    elif worker_run.version.gpu_usage == WorkerVersionGPUUsage.Required:
                        raise ValidationError(f"The task {task_name} requires a GPU and the `use_gpu` option is disabled.")
                    tasks[task_name] = task_recipe

        # Build the workflow in db
        recipe = settings.PONOS_RECIPE.copy()
        recipe['tasks'] = tasks
        if 'env' not in recipe:
            recipe['env'] = {}

        recipe['env'].update({
            'ARKINDEX_PROCESS_ID': str(self.id),
        })
        if self.mode == ProcessMode.Transkribus:
            recipe['env']['TRANSKRIBUS_EMAIL'] = settings.TRANSKRIBUS_EMAIL
            recipe['env']['TRANSKRIBUS_PASSWORD'] = settings.TRANSKRIBUS_PASSWORD
            recipe['env']['TRANSKRIBUS_WORKER_VERSION'] = str(settings.TRANSKRIBUS_WORKER_VERSION)

        elif self.mode == ProcessMode.S3:
            recipe['env']['INGEST_S3_ACCESS_KEY'] = settings.INGEST_S3_ACCESS_KEY
            recipe['env']['INGEST_S3_SECRET_KEY'] = settings.INGEST_S3_SECRET_KEY
            if settings.INGEST_S3_ENDPOINT:
                recipe['env']['INGEST_S3_ENDPOINT'] = settings.INGEST_S3_ENDPOINT
            if settings.INGEST_S3_REGION:
                recipe['env']['INGEST_S3_REGION'] = settings.INGEST_S3_REGION

        if settings.IMPORTS_WORKER_VERSION and self.mode in (ProcessMode.Files, ProcessMode.S3):
            recipe['env']['WORKER_VERSION_ID'] = str(settings.IMPORTS_WORKER_VERSION)

        if self.corpus:
            recipe['env']['ARKINDEX_CORPUS_ID'] = str(self.corpus.id)

        # Use the default Ponos farm if no farm is specified
        farm_id = farm and farm.id or get_default_farm_id()

        return Workflow.objects.create(farm_id=farm_id, recipe=yaml.dump(recipe))

    def start(self, use_cache=False, worker_activity=False, use_gpu=False, **kwargs):
        if self.mode != ProcessMode.Workers:
            assert not use_cache and not worker_activity, 'Only worker processes can be started with cache or worker activities'

        if worker_activity:
            self.activity_state = ActivityState.Pending
        if use_cache:
            self.use_cache = True
        if use_gpu:
            self.use_gpu = True

        self.workflow = self.build_workflow(**kwargs)
        self.save()
        self.workflow.start()

        # Asynchronously initialize worker activities if required
        if self.activity_state == ActivityState.Pending:
            from arkindex.project.triggers import initialize_activity
            initialize_activity(self)

        if self.mode == ProcessMode.Workers:
            CorpusWorkerVersion.objects.bulk_create([
                CorpusWorkerVersion(corpus_id=self.corpus_id, worker_version_id=worker_version_id)
                for worker_version_id in self.worker_runs.values_list('version_id', flat=True)
            ], ignore_conflicts=True)
        if self.mode == ProcessMode.Transkribus:
            worker_version_id = settings.TRANSKRIBUS_WORKER_VERSION
            CorpusWorkerVersion.objects.get_or_create(corpus_id=self.corpus_id, worker_version_id=worker_version_id)

    def retry(self):
        if self.mode == ProcessMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')

        if self.workflow:
            self.workflow.retry()

            # If worker activities are enabled for this process, no matter their previous state,
            # re-initialize the activities just in case any were deleted or stolen by other processes
            # before we retried this one.
            if self.activity_state != ActivityState.Disabled:
                from arkindex.project.triggers import initialize_activity
                initialize_activity(self)
                self.activity_state = ActivityState.Pending
                self.save()
        else:
            self.start()

    def clear(self):
        self.worker_runs.all().delete()
        self.template_id = None
        self.save()


class ProcessElement(models.Model):
    """
    Link between Processes and Elements to allow workflows from selections.
    May be removed later once the ElementPath algorithms are optimized,
    as a selection could then be turned into a set.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(Process, on_delete=models.CASCADE, related_name='process_elements')
    element = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='process_elements')

    class Meta:
        unique_together = (
            ('process', 'element'),
        )


class DataFile(S3FileMixin, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    size = models.PositiveIntegerField(help_text='file size in bytes')
    content_type = models.CharField(max_length=120)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')
    status = EnumField(S3FileStatus, default=S3FileStatus.Unchecked, max_length=50)
    trashed = models.BooleanField(default=False)

    s3_bucket = settings.AWS_STAGING_BUCKET

    @cached_property
    def s3_key(self):
        return str(self.id)

    class Meta:
        ordering = ['corpus', 'name']

    def perform_check(self, save=True, raise_exc=False):
        """
        Check Datafile's existence, content type and size from S3
        """
        try:
            assert self.exists(), 'File has not been uploaded'
            assert self.s3_object.content_length == self.size, \
                f'Uploaded file size is {self.s3_object.content_length} bytes, expected {self.size} bytes'
            assert self.s3_object.content_type == self.content_type, \
                f'Uploaded file content type is {self.s3_object.content_type}, expected {self.content_type}'
            self.status = S3FileStatus.Checked
        except AssertionError:
            self.status = S3FileStatus.Error
            if raise_exc:
                raise
        finally:
            if save:
                self.save()


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    url = models.URLField(unique=True)
    hook_token = models.CharField(max_length=250, unique=True)
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.SET_NULL, related_name='repos', blank=True, null=True)
    git_ref_revisions = models.ManyToManyField('process.Revision', through='process.GitRef')
    memberships = GenericRelation('users.Right', 'content_id')

    class Meta:
        verbose_name_plural = 'repositories'

    # Make this constant into a database field to be able to handle repositories from services other than GitLab
    provider_name = 'GitLabProvider'

    @property
    def provider_class(self):
        if not self.provider_name:
            raise ValueError("Empty provider_name")
        return get_provider(self.provider_name)

    @property
    def provider(self):
        if self.provider_class is None:
            raise NotImplementedError(f"Missing provider {self.provider_name}")
        return self.provider_class(credentials=self.credentials)

    @property
    def enabled(self):
        return self.credentials is not None and self.provider_class is not None

    def __str__(self):
        return f'{self.url}'


class GitRefType(Enum):
    Branch = 'branch'
    Tag = 'tag'


class GitRef(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    type = EnumField(GitRefType)
    name = models.CharField(max_length=250)
    repository = models.ForeignKey('process.Repository', on_delete=models.CASCADE, related_name='refs')
    revision = models.ForeignKey('process.Revision', on_delete=models.CASCADE, related_name='refs')

    class Meta:
        unique_together = (('name', 'repository'), )


class Revision(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    repo = models.ForeignKey('process.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def __str__(self):
        message = ' "{}"'.format(self.message.splitlines()[0]) if self.message else ''
        return '{}{} by {}'.format(self.hash[:8], message, self.author)


class Worker(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100)
    type = models.ForeignKey('process.WorkerType', on_delete=models.PROTECT, related_name='type')
    repository = models.ForeignKey('process.Repository', on_delete=models.CASCADE, related_name='workers')
    memberships = GenericRelation('users.Right', 'content_id')

    # Allow any authentified user to see and execute this worker
    public = models.BooleanField(default=False)

    class Meta:
        unique_together = (('slug', 'repository'),)

    def __str__(self):
        return f'{self.name}'


class WorkerType(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(max_length=100, unique=True)
    display_name = models.CharField(max_length=100)

    def __str__(self):
        return self.slug

    def __repr__(self):
        return f'<WorkerType {self.slug}>'


class WorkerVersionState(Enum):
    Created = 'created'
    Processing = 'processing'
    Available = 'available'
    Error = 'error'


class WorkerVersionGPUUsage(Enum):
    Disabled = 'disabled'
    Supported = 'supported'
    Required = 'required'


class WorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    worker = models.ForeignKey('process.Worker', on_delete=models.CASCADE, related_name='versions')
    revision = models.ForeignKey('process.Revision', on_delete=models.CASCADE, related_name='versions')
    configuration = models.JSONField()
    state = EnumField(WorkerVersionState, default=WorkerVersionState.Created)
    gpu_usage = EnumField(WorkerVersionGPUUsage, default=WorkerVersionGPUUsage.Disabled, blank=True)
    model_usage = models.BooleanField(default=False)

    # Foreign key towards the ponos Artifact holding the image content
    docker_image = models.ForeignKey(Artifact, on_delete=models.SET_NULL, null=True, blank=True)

    # The Docker internal image id (sha256:xxx) that can be shared across multiple images
    docker_image_iid = models.CharField(null=True, blank=True, max_length=80)

    corpora = models.ManyToManyField(
        'documents.Corpus',
        through='process.CorpusWorkerVersion',
        related_name='worker_versions',
    )

    class Meta:
        unique_together = (('worker', 'revision'),)
        constraints = [
            models.CheckConstraint(
                check=~models.Q(state=WorkerVersionState.Available, docker_image_id=None),
                name='workerversion_available_requires_docker_image',
            )
        ]

    def __str__(self):
        return f'{self.worker} for revision {self.revision}'

    @property
    def docker_image_name(self) -> str:
        parsed_url = urllib.parse.urlparse(self.worker.repository.url)
        return f'{parsed_url.netloc}{parsed_url.path}/{self.worker.slug}:{self.id}'.lower()

    @property
    def slug(self):
        return f'{self.worker.slug}_{self.truncated_id}'

    @property
    def truncated_id(self):
        return str(self.id)[0:6]

    @property
    def docker_command(self):
        if 'docker' not in self.configuration:
            return
        return self.configuration['docker'].get('command')

    @property
    def docker_shm_size(self):
        if 'docker' not in self.configuration:
            return
        return self.configuration['docker'].get('shm_size')


class WorkerConfiguration(IndexableModel):
    name = models.CharField(max_length=250)
    configuration = models.JSONField(default=dict)
    configuration_hash = MD5HashField()
    archived = models.BooleanField(default=False)
    worker = models.ForeignKey(
        Worker,
        on_delete=models.CASCADE,
        related_name='configurations',
    )

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<WorkerConfiguration {self.name!r} ({self.id})>'

    class Meta:
        unique_together = (
            ('worker', 'configuration_hash'),
            ('worker', 'name')
        )
        constraints = [
            models.CheckConstraint(
                check=models.Q(configuration__typeof='object'),
                name='worker_configuration_configuration_objects',
            )
        ]


class WorkerRun(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey('process.Process', on_delete=models.CASCADE, related_name='worker_runs')
    version = models.ForeignKey('process.WorkerVersion', on_delete=models.CASCADE, related_name='worker_runs')
    model_version = models.ForeignKey('training.ModelVersion', on_delete=models.CASCADE, related_name='worker_runs', null=True)
    parents = ArrayField(models.UUIDField())
    configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name='worker_runs',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    summary = models.TextField()

    class Meta:
        unique_together = (('version', 'process'),)

    def build_task_recipe(self, import_task_name, elements_path, chunk=None, workflow_runs=None):
        '''
        Build recipe dict for the Task that will represent this WorkerRun in ponos using :
        - the docker image name given by the WorkerVersion
        - the command given by the WorkerVersion
        - the artifact given by the WorkerVersion that will help us download the distant Docker image (S3)
        - a list of parent tasks slugs, using the base import task as a default
        '''
        suffix = f'_{chunk}' if chunk else ''

        if self.parents:
            if workflow_runs:
                # Use already fetched worker runs if possible
                parent_runs = [run for run in workflow_runs if run.id in self.parents]
            else:
                parent_runs = WorkerRun.objects.select_related('version').using('default').filter(id__in=self.parents)
            parents = [
                f'{worker_run.version.slug}{suffix}'
                for worker_run in parent_runs
            ]
        else:
            parents = [import_task_name]

        assert (self.version.state == WorkerVersionState.Available and self.version.docker_image_id), f"Worker Version {self.version.id} is not available and cannot be used to build a task."

        env = {
            "TASK_ELEMENTS": elements_path,
            "WORKER_VERSION_ID": str(self.version.id),
            "ARKINDEX_WORKER_RUN_ID": str(self.id),
        }
        if chunk:
            env['ARKINDEX_TASK_CHUNK'] = str(chunk)

        task_recipe = {
            'image': self.version.docker_image_iid or self.version.docker_image_name,
            'command': self.version.docker_command,
            'artifact': str(self.version.docker_image_id),
            'parents': parents,
            'env': env,
            'shm_size': self.version.docker_shm_size,
        }

        # If the worker run needs a model_version, provide its public download url as extra file
        if self.model_version:
            task_recipe["extra_files"] = {"model": settings.PUBLIC_HOSTNAME + reverse('api:model-version-download', kwargs={'pk': self.model_version.id}) + f"?token={self.model_version.build_authentication_token_hash()}"}

        return task_recipe

    def build_summary(self):
        '''
        Build a summary of the worker run displaying information about :
        - worker version used
        - model version used if provided
        - worker configuration loaded if there is one
        '''
        summary_text = f"Worker {self.version.worker.name} @ "
        git_ref_names = self.version.revision.refs.values_list('name', flat=True)
        if len(git_ref_names) > 0:
            summary_text += ", ".join(git_ref_names)
        else:
            summary_text += self.version.truncated_id

        if self.model_version:
            summary_text += f" with model {self.model_version.model.name} @ {self.model_version.truncated_id}"

        if self.configuration:
            summary_text += f" using configuration '{self.configuration.name}'"

        self.summary = summary_text


class WorkerActivityState(Enum):
    Queued = 'queued'
    Started = 'started'
    Processed = 'processed'
    Error = 'error'


class WorkerActivity(models.Model):
    """
    Many-to-many relationship between Element and WorkerVersion
    Used to track the activity of a worker version among multiple elements
    """
    # Using an UUID helps to execute SQL raw INSERT
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    created = models.DateTimeField(auto_now_add=True)

    # We will manually handle the timestamp update using a PostgreSQL trigger to avoid issues with bulk inserts/updates,
    # and we still let Django handle setting the timestamp initially, so we use `auto_now_add` and not `auto_now`.
    updated = models.DateTimeField(auto_now_add=True)

    # To properly compute a processing time, we need to separately store the time at which the state was updated to `started`.
    started = models.DateTimeField(blank=True, null=True)

    element = models.ForeignKey(
        'documents.Element',
        on_delete=models.CASCADE,
        related_name='activities',
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name='activities',
    )
    state = EnumField(
        WorkerActivityState,
        default=WorkerActivityState.Queued,
    )
    process = models.ForeignKey(
        Process,
        related_name='activities',
        on_delete=models.CASCADE,
    )
    configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name='worker_activities',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    # Specific WorkerActivity manager
    objects = ActivityManager()

    class Meta:
        constraints = [
            # Add configuration unicity
            # Two constraints are required as Null values are not compared for unicity
            models.UniqueConstraint(
                fields=['worker_version', 'element'],
                name='worker_activity_unique_no_configuration',
                condition=Q(configuration__isnull=True),
            ),
            models.UniqueConstraint(
                fields=['worker_version', 'element', 'configuration'],
                name='worker_activity_unique_configuration',
                condition=Q(configuration__isnull=False),
            ),
            # Require the `started` timestamp to be set if a WorkerActivity is in a `started` state.
            # This does not prevent activities in a processed/error state to not have start times,
            # as we do not know when activities that existed before the start time was added have started,
            # but is enough to ensure that we have to set the start time on any activity that now becomes started.
            models.CheckConstraint(
                check=~Q(state=WorkerActivityState.Started) | Q(started__isnull=False),
                name='worker_activity_started_requires_started',
            ),
        ]
        triggers = [
            pgtrigger.Trigger(
                name='update_workeractivity_updated',
                operation=pgtrigger.Update,
                when=pgtrigger.Before,
                func='NEW.updated = now(); RETURN NEW;',
            ),
            pgtrigger.ReadOnly(
                name='read_only_workeractivity_updated',
                fields=['updated'],
            )
        ]


class CorpusWorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    corpus = models.ForeignKey(
        'documents.Corpus',
        on_delete=models.CASCADE,
        related_name='worker_version_cache',
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name='corpus_cache',
    )

    objects = CorpusWorkerVersionManager()

    class Meta:
        unique_together = (
            ('corpus', 'worker_version')
        )
