import uuid
from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.process.models import GitRefType, ProcessMode, Revision, WorkerRun, WorkerVersion, WorkerVersionState
from arkindex.process.utils import get_default_farm_id
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role
from ponos.models import State, Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''

# Mock WorkerRun versions docker_image_id attribute
ARTIFACT_ID = uuid.uuid4()


class TestWorkerRuns(FixtureAPITestCase):
    """
    Test worker runs endpoints and methods
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process_1 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_1 = cls.version_1.worker
        cls.repo = cls.worker_1.repository
        cls.run_1 = cls.process_1.worker_runs.create(version=cls.version_1, parents=[])
        cls.configuration_1 = cls.worker_1.configurations.create(name="My config", configuration={"key": "value"})
        worker_version = WorkerVersion.objects.exclude(worker=cls.version_1.worker).first()
        cls.configuration_2 = worker_version.worker.configurations.create(name="Config")
        cls.process_2 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)
        # Add an execution access right on the worker
        cls.worker_1.memberships.create(user=cls.user, level=Role.Contributor.value)

        cls.creds = cls.user.credentials.get()
        cls.gl_patch = patch('arkindex.process.providers.Gitlab')

        # Model and Model version setup
        cls.model_1 = Model.objects.create(name='My model')
        cls.model_1.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.model_version_1 = ModelVersion.objects.create(model=cls.model_1, state=ModelVersionState.Available, size=8, hash='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa', archive_hash='bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb')

    def test_runs_list_requires_login(self):
        response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.process_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_runs_list_no_execution_right(self):
        """
        Worker runs attached to a process can be listed even if the user has no execution rights to workers
        This is due to the fact a user can see a process running on a corpus they have access
        """
        self.worker_1.memberships.all().delete()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.process_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 1)

    def test_runs_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.process_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['results'], [{
            'id': str(self.run_1.id),
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
            'parents': [],
            'model_version': None,
            'configuration': None,
            'process':{
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
        }])

    def test_runs_list_filter_process(self):
        run_2 = self.process_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(run_2.id))

    def test_runs_post_requires_login(self):
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_runs_post_no_execution_right(self):
        """
        An execution access on the target worker version is required to create a worker run
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version_id': ['You do not have an execution access to this version.']})

    def test_runs_post_invalid_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
            data={'worker_version_id': uuid.uuid4(), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version_id': ['This version does not exist.']})

    def test_create_run_unique(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_1.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version_id': ['A WorkerRun with this version already exists on this process.']})

    def test_create_run_unavailable_version(self):
        self.version_1.state = WorkerVersionState.Error
        self.version_1.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version_id': ['This worker version is not in an Available state.']})

    def test_runs_post_invalid_process_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': uuid.uuid4()}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_runs_post_invalid_process_mode(self):
        img_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Files)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(img_process.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'process': ['Import mode must be Workers']})

    def test_runs_post_process_already_started(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            workflow=Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE),
        )

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(process.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'__all__': [
            'Cannot create a WorkerRun on a Process that has already started'
        ]})

    def test_runs_post_create_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
                data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        pk = data.pop('id')
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
            'parents': [],
            'model_version': None,
            'configuration': None,
            'process': {
                'id': str(self.process_2.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None,
            },
        })
        run = WorkerRun.objects.get(pk=pk)
        # Check generated summary, this supposes that self.version_1.revision does not have any git refs
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(self.version_1.id)[0:6]}")

    def test_runs_post_via_repository_right(self):
        self.worker_1.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.worker_1.repository.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
                data={'worker_version_id': str(self.version_1.id), 'parents': []}, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_runs_post_empty(self):
        """
        Parents and worker_version_id are required to create a worker run
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'parents': ['This field is required.'],
            'worker_version_id': ['This field is required.']
        })

    def test_create_run_configuration(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(20):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
                data={
                    'worker_version_id': str(self.version_1.id),
                    'parents': [],
                    'configuration_id': str(self.configuration_1.id)
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop('id')
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
            'parents': [],
            'model_version': None,
            'configuration': {
                'id': str(self.configuration_1.id),
                'archived': False,
                'configuration': {'key': 'value'},
                'name': 'My config'
            },
            'process': {
                'id': str(self.process_2.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
        })
        run = WorkerRun.objects.get(pk=pk)
        # Check generated summary, this supposes that self.version_1.revision does not have any git refs
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(self.version_1.id)[0:6]} using configuration '{self.configuration_1.name}'")

    def test_create_run_invalid_configuration(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
            data={'worker_version_id': str(self.version_1.id), 'parents': [], 'configuration_id': str(self.configuration_2.id)}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'configuration_id': ['The configuration must be part of the same worker.']})

    def test_runs_worker_run_summary_with_git_refs(self):
        """
        Create a worker run with a worker version related to a revision tagged with a GitRef
        """
        self.client.force_login(self.user)
        # Version with no git refs
        revision = Revision.objects.create(
            repo_id=self.repo.id,
            hash=uuid.uuid4().hex,
            message="Fake revision",
            author="Teklia Bot"
        )
        git_ref_names = ["main", "develop", "trunk", "master", "patate", "pouet"]
        expected_refs = []
        for git_ref_name in git_ref_names:
            new_ref = revision.refs.create(
                name=git_ref_name,
                type=GitRefType.Branch,
                repository=revision.repo
            )
            expected_refs.append({'id': str(new_ref.id), 'name': git_ref_name, 'type': 'branch'})

        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=revision,
            configuration={},
            state=WorkerVersionState.Available,
            docker_image_id=self.version_1.docker_image_id,
        )

        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:worker-run-list', kwargs={'pk': str(self.process_2.id)}),
                data={'worker_version_id': str(version.id), 'parents': []}, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        pk = data.pop('id')
        self.assertNotEqual(pk, self.run_1.id)
        self.assertDictEqual(data, {
            'worker_version': {
                'id': str(version.id),
                'configuration': {},
                'docker_image': str(self.version_1.docker_image_id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{version.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(version.revision.id),
                    'author': 'Teklia Bot',
                    'commit_url': f'http://my_repo.fake/workers/worker/commit/{version.revision.hash}',
                    'created': version.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': version.revision.hash,
                    'message': 'Fake revision',
                    'refs': expected_refs,
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
            'parents': [],
            'model_version': None,
            'configuration': None,
            'process': {
                'id': str(self.process_2.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None,
            },
        })
        run = WorkerRun.objects.get(pk=pk)
        # Check generated summary
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ main, develop, trunk, master, patate, pouet")

    def test_retrieve_run_requires_login(self):
        response = self.client.get(reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_run_no_worker_execution_right(self):
        """
        A user can retrieve any worker run if they have an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })

    def test_retrieve_run_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })

    @override_settings(PUBLIC_HOSTNAME='https://arkindex.localhost')
    def test_gitrefs_are_retrieved(self):
        """
        Check the GitRefs are retrieved with the revision
        """
        from arkindex.process.providers import GitLabProvider

        # Create gitrefs and check they were created
        self.gl_mock = self.gl_patch.start()
        commit_refs = [
            {'name': 'refs/tags/v0.1.0', 'type': 'tag'},
            {'name': 'refs/heads/branch1', 'type': 'branch'},
            {'name': 'refs/heads/branch2', 'type': 'branch'},
        ]
        revision = Revision.objects.create(
            repo=self.repo,
            hash='1',
            message='commit message',
            author='bob',
        )

        for ref in commit_refs:
            GitLabProvider(credentials=self.creds) \
                .update_or_create_ref(self.repo, revision, ref['name'], ref['type'])

        refs = [
            {'name': ref.name, 'type': ref.type, 'repo': ref.repository}
            for ref in revision.refs.all()
        ]
        self.assertListEqual(refs, [
            {'name': 'refs/tags/v0.1.0', 'type': GitRefType.Tag, 'repo': self.repo},
            {'name': 'refs/heads/branch1', 'type': GitRefType.Branch, 'repo': self.repo},
            {'name': 'refs/heads/branch2', 'type': GitRefType.Branch, 'repo': self.repo},
        ])
        self.assertEqual(self.repo.refs.count(), 3)

        # Assign the revision with gitrefs to the worker version
        self.version_1.revision = revision
        self.version_1.save()
        self.assertTrue(revision.refs.exists())

        # Check that the gitrefs are retrieved with RetrieveWorkerRun
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        version_revision = response.json()['worker_version']['revision']
        refs = version_revision.pop('refs')
        # Only asserting on the name and type of the refs since they were created in bulk using
        # GitLabProvider.update_or_create_ref, ignoring the IDs.
        refs_no_id = [
            {
                "name": item["name"],
                "type": item["type"]
            }
            for item in refs
        ]
        self.assertDictEqual(version_revision, {
            "author": "bob",
            "commit_url": "http://my_repo.fake/workers/worker/commit/1",
            "created": revision.created.isoformat().replace('+00:00', 'Z'),
            "hash": "1",
            "id": str(revision.id),
            "message": "commit message",
        })
        self.assertCountEqual(refs_no_id, [
            {
                "name": "refs/tags/v0.1.0",
                "type": "tag"
            },
            {
                "name": "refs/heads/branch1",
                "type": "branch"
            },
            {
                "name": "refs/heads/branch2",
                "type": "branch"
            }
        ])

    def test_update_run_requires_id_and_parents(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'configuration_id': str(self.configuration_1.id),
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'parents': ['This field is required.']})

    def test_update_run_requires_login(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        with self.assertNumQueries(0):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_run_no_project_admin_right(self):
        """
        A user cannot update a worker run if he has no admin access on its process project
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': []
                }
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have an admin access to the process project.'})

    def test_update_run_invalid_id(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_run_inexistant_parent(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': ['12341234-1234-1234-1234-123412341234'],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), [
            f"Can't add or update WorkerRun {self.run_1.id} because parents field isn't properly defined. It can be either because"
            " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
            " same Process than this WorkerRun."
        ])

    def test_update_run_process_id(self):
        """
        Process field cannot be updated
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'process_id': str(self.process_2.id),
                    'parents': [],
                }, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.process.id, self.process_1.id)

    def test_update_run_worker_version_id(self):
        """
        Version field cannot be updated
        """
        self.client.force_login(self.user)
        dla_version = WorkerVersion.objects.get(worker__slug='dla')
        with self.assertNumQueries(10):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'worker_version_id': str(dla_version.id),
                    'parents': [],
                }, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.run_1.refresh_from_db()
        self.assertNotEqual(self.run_1.version_id, dla_version.id)

    def test_update_run_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        # Check generated summary, before updating, it should not be that verbose
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]}")
        with self.assertNumQueries(11):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': self.run_1.id}),
                data={
                    'parents': [],
                    'configuration_id': str(self.configuration_1.id)
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.run_1.refresh_from_db()

        self.assertEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': {
                'id': str(self.configuration_1.id),
                'archived': False,
                'configuration': {'key': 'value'},
                'name': 'My config'
            },
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.assertEqual(self.run_1.configuration.id, self.configuration_1.id)
        # Check generated summary, after the update, the configuration should be displayed as well
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]} using configuration '{self.configuration_1.name}'")

    def test_update_run_invalid_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': self.run_1.id}),
                data={'parents': [], 'configuration_id': str(self.configuration_2.id)},
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'configuration_id': ['The configuration must be part of the same worker.']})

    def test_update_run_process_already_started(self):
        """
        Update dependencies of a worker run is not possible once the process is started
        """
        self.process_1.start()
        self.assertIsNotNone(self.process_1.workflow)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            '__all__': ['Cannot update a WorkerRun on a Process that has already started']
        })

    def test_update_run_model_version_not_allowed(self):
        """
        The model_version UUID is not allowed when the related version doesn't allow model_usage
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=False
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(run_2.id)}),
                data={
                    'model_version_id': str(self.model_version_1.id),
                    'parents': []
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'model_version_id': ['This worker version does not support model usage.']
        })

    def test_update_run_unknown_model_version(self):
        """
        Cannot use a model version id that doesn't exist
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=True
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        random_model_version_uuid = str(uuid.uuid4())
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(run_2.id)}),
                data={
                    'model_version_id': random_model_version_uuid,
                    'parents': []
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'model_version_id': [f'Invalid pk "{random_model_version_uuid}" - object does not exist.']
        })

    def test_update_run_model_version_no_access(self):
        """
        Cannot update a worker run with a model_version UUID, when you don't have access to the model version
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=True
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        # Create a model version, the user has no access to
        model_no_access = Model.objects.create(name='Secret model')
        model_version_no_access = ModelVersion.objects.create(model=model_no_access, state=ModelVersionState.Available, size=8, hash='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa', archive_hash='bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb')

        with self.assertNumQueries(10):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(run_2.id)}),
                data={
                    'model_version_id': str(model_version_no_access.id),
                    'parents': []
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'model_version_id': ['You do not have access to this model version.']
        })

    def test_update_run_model_version(self):
        """
        Update the worker run by adding a model_version with a worker version that supports it
        """
        self.client.force_login(self.user)
        rev = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev,
            configuration={"test": "test2"},
            model_usage=True
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertEqual(run.model_version, None)
        # Check generated summary, before updating, there should be only information about the worker version
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]}")
        with self.assertNumQueries(15):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(run.id)}),
                data={
                    'model_version_id': str(self.model_version_1.id),
                    'parents': []
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        run.refresh_from_db()

        self.assertEqual(response.json(), {
            'id': str(run.id),
            'configuration': None,
            'model_version': {
                'id': str(self.model_version_1.id),
                'configuration': {},
                'model': {
                    'id': str(self.model_1.id),
                    'name': 'My model'
                },
                'size': 8,
                'state': 'available',
                'tag': None
            },
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(version_with_model.id),
                'configuration': {'test': 'test2'},
                'docker_image': None,
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{version_with_model.id}',
                'gpu_usage': 'disabled',
                'model_usage': True,
                'revision': {
                    'id': str(rev.id),
                    'author': 'bob',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/2',
                    'created': rev.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '2',
                    'message': 'beep boop',
                    'refs': []
                },
                'state': 'created',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        # Check generated summary, after updating, there should be information about the model loaded
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]} with model {self.model_version_1.model.name} @ {str(self.model_version_1.id)[:6]}")

    def test_update_run_configuration_and_model_version(self):
        """
        Update the worker run by adding both a model_version and a worker configuration
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=True
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertIsNone(run.model_version)
        self.assertIsNone(run.configuration)
        # Check generated summary, before updating, there should be only information about the worker version
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]}")
        with self.assertNumQueries(16):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(run.id)}),
                data={
                    'model_version_id': str(self.model_version_1.id),
                    'configuration_id': str(self.configuration_1.id),
                    'parents': []
                }, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        run.refresh_from_db()
        self.assertEqual(response.json(), {
            'id': str(run.id),
            'configuration': {
                'id': str(self.configuration_1.id),
                'archived': False,
                'configuration': {'key': 'value'},
                'name': 'My config'
            },
            'model_version': {
                'id': str(self.model_version_1.id),
                'configuration': {},
                'model': {
                    'id': str(self.model_1.id),
                    'name': 'My model'
                },
                'size': 8,
                'state': 'available',
                'tag': None
            },
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(version_with_model.id),
                'configuration': {'test': 'test2'},
                'docker_image': None,
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{version_with_model.id}',
                'gpu_usage': 'disabled',
                'model_usage': True,
                'revision': {
                    'id': str(rev_2.id),
                    'author': 'bob',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/2',
                    'created': rev_2.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '2',
                    'message': 'beep boop',
                    'refs': []
                },
                'state': 'created',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        # Check generated summary, after updating, there should be information about the model loaded
        self.assertEqual(run.summary, f"Worker {self.worker_1.name} @ {str(version_with_model.id)[:6]} with model {self.model_version_1.model.name} @ {str(self.model_version_1.id)[:6]} using configuration '{self.configuration_1.name}'")

    def test_update_run(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(12):
            response = self.client.put(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.run_1.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [str(run_2.id)],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })

    def test_partial_update_run_requires_login(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_run_no_project_admin_right(self):
        """
        A user cannot update a worker run if he has no admin access on its process project
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have an admin access to the process project.'})

    def test_partial_update_run_invalid_id(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_partial_update_run_inexistant_parent(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': ['12341234-1234-1234-1234-123412341234'],
                }, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), [
            f"Can't add or update WorkerRun {self.run_1.id} because parents field isn't properly defined. It can be either because"
            " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
            " same Process than this WorkerRun."
        ])

    def test_partial_update_run_process_id(self):
        """
        Process field cannot be updated
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'process_id': str(self.process_2.id),
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.process.id, self.process_1.id)

    def test_partial_update_run_worker_version_id(self):
        """
        Version field cannot be updated
        """
        self.client.force_login(self.user)
        dla_version = WorkerVersion.objects.get(worker__slug='dla')
        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'worker_version_id': str(dla_version.id),
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.run_1.refresh_from_db()
        self.assertNotEqual(self.run_1.version_id, dla_version.id)

    def test_partial_update_run_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]}")
        with self.assertNumQueries(11):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': self.run_1.id}),
                data={
                    'configuration_id': str(self.configuration_1.id)
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.run_1.refresh_from_db()
        self.assertEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': {
                'id': str(self.configuration_1.id),
                'archived': False,
                'configuration': {'key': 'value'},
                'name': 'My config'
            },
            'model_version': None,
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.assertEqual(self.run_1.configuration.id, self.configuration_1.id)
        self.assertEqual(self.run_1.summary, f"Worker {self.run_1.version.worker.name} @ {str(self.run_1.version.id)[:6]} using configuration '{self.configuration_1.name}'")

    def test_partial_update_run_invalid_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': self.run_1.id}),
                data={'configuration_id': str(self.configuration_2.id)},
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'configuration_id': ['The configuration must be part of the same worker.']})

    def test_partial_update_run_process_already_started(self):
        """
        Update dependencies of a worker run is not possible once the process is started
        """
        self.process_1.start()
        self.assertIsNotNone(self.process_1.workflow)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            '__all__': ['Cannot update a WorkerRun on a Process that has already started']
        })

    def test_partial_update_run_model_version_not_allowed(self):
        """
        The model_version UUID is not allowed when the related version doesn't allow model_usage
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=False
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(run_2.id)}),
                data={
                    'model_version_id': str(self.model_version_1.id),
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'model_version_id': ['This worker version does not support model usage.']
        })

    def test_partial_update_run_unknown_model_version(self):
        """
        Cannot use a model version id that doesn't exist
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=True
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        random_model_version_uuid = str(uuid.uuid4())
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(run_2.id)}),
                data={
                    'model_version_id': random_model_version_uuid,
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'model_version_id': [f'Invalid pk "{random_model_version_uuid}" - object does not exist.']
        })

    def test_partial_update_run_model_version_no_access(self):
        """
        Cannot update a worker run with a model_version UUID, when you don't have access to the model version
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=True
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        # Create a model version, the user has no access to
        model_no_access = Model.objects.create(name='Secret model')
        model_version_no_access = ModelVersion.objects.create(model=model_no_access, state=ModelVersionState.Available, size=8, hash='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa', archive_hash='bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb')

        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(run_2.id)}),
                data={
                    'model_version_id': str(model_version_no_access.id),
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'model_version_id': ['You do not have access to this model version.']
        })

    def test_partial_update_run_model_version(self):
        """
        Update the worker run by adding a model_version with a worker version that supports it
        """
        self.client.force_login(self.user)
        rev = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev,
            configuration={"test": "test2"},
            model_usage=True
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertIsNone(run.model_version_id)
        self.assertEqual(run.summary, f"Worker {version_with_model.worker.name} @ {str(version_with_model.id)[:6]}")
        with self.assertNumQueries(15):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(run.id)}),
                data={
                    'model_version_id': str(self.model_version_1.id),
                }, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        run.refresh_from_db()
        self.assertEqual(response.json(), {
            'id': str(run.id),
            'configuration': None,
            'model_version': {
                'id': str(self.model_version_1.id),
                'configuration': {},
                'model': {
                    'id': str(self.model_1.id),
                    'name': 'My model'
                },
                'size': 8,
                'state': 'available',
                'tag': None
            },
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(version_with_model.id),
                'configuration': {'test': 'test2'},
                'docker_image': None,
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{version_with_model.id}',
                'gpu_usage': 'disabled',
                'model_usage': True,
                'revision': {
                    'id': str(rev.id),
                    'author': 'bob',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/2',
                    'created': rev.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '2',
                    'message': 'beep boop',
                    'refs': []
                },
                'state': 'created',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        self.assertEqual(run.summary, f"Worker {version_with_model.worker.name} @ {str(version_with_model.id)[:6]} with model {self.model_version_1.model.name} @ {str(self.model_version_1.id)[:6]}")

    def test_partial_update_run_model_version_with_configuration(self):
        """
        Updating the worker run by adding a model_version with a worker version
        doesn't erase previously loaded worker configuration
        """
        self.client.force_login(self.user)
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='beep boop',
            author='bob',
        )
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"},
            model_usage=True
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
            configuration=self.configuration_1
        )
        self.assertEqual(run.model_version_id, None)
        with self.assertNumQueries(15):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(run.id)}),
                data={
                    'model_version_id': str(self.model_version_1.id),
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        run.refresh_from_db()
        self.assertEqual(response.json(), {
            'id': str(run.id),
            'configuration': {
                'archived': False,
                'configuration': {'key': 'value'},
                'id': str(self.configuration_1.id),
                'name': 'My config'
            },
            'model_version': {
                'id': str(self.model_version_1.id),
                'configuration': {},
                'model': {
                    'id': str(self.model_1.id),
                    'name': 'My model'
                },
                'size': 8,
                'state': 'available',
                'tag': None
            },
            'parents': [],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(version_with_model.id),
                'configuration': {'test': 'test2'},
                'docker_image': None,
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{version_with_model.id}',
                'gpu_usage': 'disabled',
                'model_usage': True,
                'revision': {
                    'id': str(rev_2.id),
                    'author': 'bob',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/2',
                    'created': rev_2.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '2',
                    'message': 'beep boop',
                    'refs': []
                },
                'state': 'created',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        self.assertEqual(run.configuration_id, self.configuration_1.id)

    def test_partial_update_run(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(12):
            response = self.client.patch(
                reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)}),
                data={
                    'parents': [str(run_2.id)],
                }, format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.run_1.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(self.run_1.id),
            'configuration': None,
            'model_version': None,
            'parents': [str(run_2.id)],
            'process': {
                'id': str(self.process_1.id),
                'activity_state': 'disabled',
                'corpus': str(self.corpus.id),
                'mode': 'workers',
                'model_id': None,
                'name': None,
                'state': 'unscheduled',
                'test_folder_id': None,
                'train_folder_id': None,
                'validation_folder_id': None,
                'workflow': None
            },
            'worker_version': {
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'docker_image': str(self.version_1.docker_image.id),
                'docker_image_iid': None,
                'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                'gpu_usage': 'disabled',
                'model_usage': False,
                'revision': {
                    'id': str(self.version_1.revision.id),
                    'author': 'Test user',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'created': self.version_1.revision.created.isoformat().replace('+00:00', 'Z'),
                    'hash': '1337',
                    'message': 'My w0rk3r',
                    'refs': []
                },
                'state': 'available',
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer'
                }
            },
        })

    def test_delete_run_requires_login(self):
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_run_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_run_no_worker_execution_right(self):
        """
        A user can delete a worker run with no right on its worker but an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        self.assertEqual(WorkerRun.objects.count(), 3)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(WorkerRun.objects.count(), 2)
        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()

    def test_delete_run(self):
        self.client.force_login(self.user)
        self.assertEqual(WorkerRun.objects.count(), 3)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(WorkerRun.objects.count(), 2)
        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()

    def test_delete_run_with_children(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        rev_3 = self.repo.revisions.create(
            hash='3',
            message='commit message',
            author='dylan',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.process_1.worker_runs.create(
            version=version_3,
            parents=[self.run_1.id, run_2.id],
        )

        self.assertEqual(WorkerRun.objects.count(), 5)
        self.assertTrue(self.run_1.id in run_2.parents)
        self.assertTrue(self.run_1.id in run_3.parents)
        self.client.force_login(self.user)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(len(WorkerRun.objects.all()), 4)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()
        run_2.refresh_from_db()
        run_3.refresh_from_db()
        self.assertEqual(run_2.parents, [])
        self.assertEqual(run_3.parents, [run_2.id])

    def test_delete_run_started_process(self):
        """
        A user shouldn't be able to delete the worker run of a started process
        """
        self.client.force_login(self.user)
        self.process_1.start()
        self.process_1.workflow.tasks.update(state=State.Running)
        response = self.client.delete(
            reverse('api:worker-run-details', kwargs={'pk': str(self.run_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'__all__': ["Cannot update a WorkerRun on a Process that has already started"]})

    def test_build_task_recipe_no_parent(self):
        self.version_1.docker_image_id = ARTIFACT_ID
        self.assertDictEqual(self.run_1.build_task_recipe('import', '/data/import/elements.json'), {
            'image': f'my_repo.fake/workers/worker/reco:{str(self.version_1.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': ['import'],
            'shm_size': None,
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(self.version_1.id),
                'ARKINDEX_WORKER_RUN_ID': str(self.run_1.id),
            },
        })

    def test_build_task_recipe_with_chunk(self):
        self.version_1.docker_image_id = ARTIFACT_ID
        self.assertDictEqual(self.run_1.build_task_recipe('import', '/data/import/elements.json', chunk=4), {
            'image': f'my_repo.fake/workers/worker/reco:{str(self.version_1.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': ['import'],
            'shm_size': None,
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(self.version_1.id),
                'ARKINDEX_WORKER_RUN_ID': str(self.run_1.id),
                'ARKINDEX_TASK_CHUNK': '4',
            },
        })

    def test_build_task_recipe_with_parent(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_id = ARTIFACT_ID
        version_2.state = WorkerVersionState.Available
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        self.assertDictEqual(run_2.build_task_recipe('import', '/data/import/elements.json'), {
            'image': f'my_repo.fake/workers/worker/reco:{str(version_2.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': [f'reco_{str(self.version_1.id)[0:6]}'],
            'shm_size': None,
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(version_2.id),
                'ARKINDEX_WORKER_RUN_ID': str(run_2.id),
            },
        })

    def test_build_task_recipe_with_parent_and_chunk(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_id = ARTIFACT_ID
        version_2.state = WorkerVersionState.Available
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        self.assertDictEqual(run_2.build_task_recipe('import', '/data/import/elements.json', chunk=4), {
            'image': f'my_repo.fake/workers/worker/reco:{str(version_2.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': [f'reco_{str(self.version_1.id)[0:6]}_4'],
            'shm_size': None,
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(version_2.id),
                'ARKINDEX_WORKER_RUN_ID': str(run_2.id),
                'ARKINDEX_TASK_CHUNK': '4',
            },
        })

    def test_build_task_recipe_shm_size(self):
        self.version_1.docker_image_id = ARTIFACT_ID
        self.version_1.configuration = {
            'docker': {
                'shm_size': 505,
            }
        }
        self.assertDictEqual(self.run_1.build_task_recipe('import', '/data/import/elements.json', chunk=4), {
            'image': f'my_repo.fake/workers/worker/reco:{str(self.version_1.id)}',
            'command': None,
            'artifact': str(ARTIFACT_ID),
            'parents': ['import'],
            'shm_size': 505,
            'env': {
                'TASK_ELEMENTS': '/data/import/elements.json',
                'WORKER_VERSION_ID': str(self.version_1.id),
                'ARKINDEX_WORKER_RUN_ID': str(self.run_1.id),
                'ARKINDEX_TASK_CHUNK': '4',
            },
        })

    def test_build_task_recipe_unavailable_version(self):
        rev_2 = self.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_2,
            configuration={"test": "test2"}
        )
        version_2.docker_image_id = ARTIFACT_ID
        self.assertEqual(version_2.state, WorkerVersionState.Created)
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )

        with self.assertRaisesRegex(
            AssertionError,
            f"Worker Version {version_2.id} is not available and cannot be used to build a task."
        ):
            run_2.build_task_recipe('import', '/data/import/elements.json')
